"""OpenFOAM module of the API for fluid dynamics."""
import pathlib
from typing import Optional, List
from uuid import UUID

from inductiva import types
from inductiva.simulation import Simulator


class OpenFOAM(Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    def __init__(self, api_method: str = "fvm.openfoam.run_simulation"):
        super().__init__()
        self.api_method = api_method

    @property
    def api_method_name(self) -> str:
        return self.api_method

    def run(
        self,
        input_dir: types.Path,
        commands: List[dict],
        output_dir: Optional[types.Path] = None,
        resource_pool_id: Optional[UUID] = None,
        n_cores: int = 1,
    ) -> pathlib.Path:
        """Run the simulation.

        Args:
            commands: List of commands to run using the OpenFOAM simulator.
            n_cores: Number of MPI cores to use for the simulation.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            output_dir=output_dir,
            resource_pool_id=resource_pool_id,
            n_cores=n_cores,
            commands=commands,
        )

    def run_async(
        self,
        input_dir: types.Path,
        commands: List[dict],
        resource_pool_id: Optional[UUID] = None,
        n_cores: int = 1,
    ) -> str:
        """Run the simulation asynchronously.

        Args:
            n_cores: Number of MPI cores to use for the simulation.
            method_name: OpenFOAM method to run. This involves commands
                from pre-processing, to solvers and post-processing.
            openfoam_flags: Flags to pass to the openfoam method_name.
            other arguments: See the documentation of the base class.
            """

        return super().run_async(input_dir,
                                 resource_pool_id=resource_pool_id,
                                 commands=commands,
                                 n_cores=n_cores)



"""Physical properties of different fluid types."""

from dataclasses import dataclass


@dataclass(eq=True, frozen=True)
class FluidType:
    density: float
    kinematic_viscosity: float


WATER = FluidType(
    density=1e3,
    kinematic_viscosity=1e-6,
)

HONEY = FluidType(
    density=1360,
    kinematic_viscosity=7.36e-5,
)

OLIVE_OIL = FluidType(
    density=905,
    kinematic_viscosity=4.32e-5,
)

LIQUID_PROPANE = FluidType(
    density=580,
    kinematic_viscosity=2.73e-7,
)

JET_FUEL = FluidType(
    density=802.5,
    kinematic_viscosity=7.6e-6,
)

BEER = FluidType(
    density=1.007e3,
    kinematic_viscosity=1.8e-6,
)

GEAR_OIL = FluidType(
    density=860,
    kinematic_viscosity=4.2e-6,
)


def get_fluid_color(fluid):
    colors = {
        WATER: "blue",
        HONEY: "darkgoldenrod",
        OLIVE_OIL: "olivedrab",
        BEER: "goldenrod",
        GEAR_OIL: "peru",
        JET_FUEL: "darkgoldenrod",
        LIQUID_PROPANE: "blueviolet"
    }

    return colors[fluid]



"""Methods for interacting with the admin APIs."""
from .tasks import get_tasks_info
from .executers import (create_resource_pool, launch_executer, kill_executer,
                        launch_executer_group, kill_executer_group)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ExecuterCreate(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Info for creating an executer.
    """

    class MetaOapg:
        required = {
            "memory",
            "cpu_info",
            "create_time",
            "cpu_count_logical",
            "cpu_count_physical",
            "executer_type",
            "host_info",
        }

        class properties:
            create_time = schemas.DateTimeSchema
            executer_type = schemas.StrSchema
            cpu_count_logical = schemas.IntSchema
            cpu_count_physical = schemas.IntSchema
            memory = schemas.IntSchema
            cpu_info = schemas.StrSchema

            class host_info(
                    schemas.ComposedSchema,):

                class MetaOapg:

                    @classmethod
                    @functools.lru_cache()
                    def one_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            GCloudHostInfo,
                            InductivaHostInfo,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'host_info':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            resource_pool_id = schemas.StrSchema
            __annotations__ = {
                "create_time": create_time,
                "executer_type": executer_type,
                "cpu_count_logical": cpu_count_logical,
                "cpu_count_physical": cpu_count_physical,
                "memory": memory,
                "cpu_info": cpu_info,
                "host_info": host_info,
                "resource_pool_id": resource_pool_id,
            }

    memory: MetaOapg.properties.memory
    cpu_info: MetaOapg.properties.cpu_info
    create_time: MetaOapg.properties.create_time
    cpu_count_logical: MetaOapg.properties.cpu_count_logical
    cpu_count_physical: MetaOapg.properties.cpu_count_physical
    executer_type: MetaOapg.properties.executer_type
    host_info: MetaOapg.properties.host_info

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["create_time"]
    ) -> MetaOapg.properties.create_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["executer_type"]
    ) -> MetaOapg.properties.executer_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["memory"]
    ) -> MetaOapg.properties.memory:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_info"]
    ) -> MetaOapg.properties.cpu_info:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["host_info"]
    ) -> MetaOapg.properties.host_info:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["resource_pool_id"]
    ) -> MetaOapg.properties.resource_pool_id:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "create_time",
        "executer_type",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory",
        "cpu_info",
        "host_info",
        "resource_pool_id",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["create_time"]
    ) -> MetaOapg.properties.create_time:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["executer_type"]
    ) -> MetaOapg.properties.executer_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["memory"]
    ) -> MetaOapg.properties.memory:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_info"]
    ) -> MetaOapg.properties.cpu_info:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["host_info"]
    ) -> MetaOapg.properties.host_info:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["resource_pool_id"]
    ) -> typing.Union[MetaOapg.properties.resource_pool_id, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "create_time",
        "executer_type",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory",
        "cpu_info",
        "host_info",
        "resource_pool_id",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        memory: typing.Union[
            MetaOapg.properties.memory,
            decimal.Decimal,
            int,
        ],
        cpu_info: typing.Union[
            MetaOapg.properties.cpu_info,
            str,
        ],
        create_time: typing.Union[
            MetaOapg.properties.create_time,
            str,
            datetime,
        ],
        cpu_count_logical: typing.Union[
            MetaOapg.properties.cpu_count_logical,
            decimal.Decimal,
            int,
        ],
        cpu_count_physical: typing.Union[
            MetaOapg.properties.cpu_count_physical,
            decimal.Decimal,
            int,
        ],
        executer_type: typing.Union[
            MetaOapg.properties.executer_type,
            str,
        ],
        host_info: typing.Union[
            MetaOapg.properties.host_info,
            dict,
            frozendict.frozendict,
            str,
            date,
            datetime,
            uuid.UUID,
            int,
            float,
            decimal.Decimal,
            bool,
            None,
            list,
            tuple,
            bytes,
            io.FileIO,
            io.BufferedReader,
        ],
        resource_pool_id: typing.Union[MetaOapg.properties.resource_pool_id,
                                       str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'ExecuterCreate':
        return super().__new__(
            cls,
            *_args,
            memory=memory,
            cpu_info=cpu_info,
            create_time=create_time,
            cpu_count_logical=cpu_count_logical,
            cpu_count_physical=cpu_count_physical,
            executer_type=executer_type,
            host_info=host_info,
            resource_pool_id=resource_pool_id,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.g_cloud_host_info import GCloudHostInfo
from inductiva.client.model.inductiva_host_info import InductivaHostInfo



"""Test files module."""
import glob
import pathlib
from .files import get_timestamped_path


def test_get_timestamped_path_with_ext(tmp_path: pathlib.Path):
    """Check if files with extensions are correctly created."""
    path = tmp_path / "file.txt"
    path.touch()

    timestamped_path = get_timestamped_path(path)

    assert not timestamped_path.exists()
    assert timestamped_path.name.startswith("file-")
    assert timestamped_path.name.endswith(".txt")


def test_get_timestamp_path(tmp_path: pathlib.Path):
    """Check if multiple dirs are created with different names."""
    n = 10

    path = tmp_path / "output"

    for _ in range(n):
        path = get_timestamped_path(path)
        path.mkdir()

    assert len(glob.glob(str(path.parent / "output-*"))) == n



"""Base class for scenarios."""

from abc import ABC, abstractmethod
import tempfile
from typing import Optional
from uuid import UUID
from inductiva.types import Path
from inductiva.simulation import Simulator
from inductiva.utils.files import resolve_path, get_timestamped_path
from inductiva.utils.misc import split_camel_case
from inductiva.tasks.methods import get_task_info, fetch_task_output
import json


class Scenario(ABC):
    """Base class for scenarios."""

    def _setup_output_dir(self, output_dir: str):
        """Setup the scenario output directory."""
        if output_dir is None:
            scenario_name_splitted = split_camel_case(self.__class__.__name__)
            output_dir_prefix = "-".join(scenario_name_splitted).lower()
            output_dir = get_timestamped_path(f"{output_dir_prefix}-output")
        output_dir = resolve_path(output_dir)
        return output_dir

    def _setup_config(self, simulator: Simulator, input_dir: Path):
        """Setup the scenario configuration files and arguments."""
        self.gen_aux_files(simulator, input_dir)
        self.gen_config(simulator, input_dir)

        args = ()
        config_filename = self.get_config_filename(simulator)
        if config_filename:
            args += (config_filename,)
        return args

    def download_outputs(self, output_dir: str = None):
        """Download the outputs of an async simulation to output_dir."""

        if not get_task_info(self.task_id)["status"]:
            raise ValueError("Simulation not finished.")

        fetch_task_output(self.task_id, output_dir=output_dir, return_type=None)

    @abstractmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: Path):
        """To be implemented in subclasses."""
        pass

    @abstractmethod
    def gen_config(self, simulator: Simulator, input_dir: Path):
        """To be implemented in subclasses."""
        pass

    @abstractmethod
    def get_config_filename(self, simulator: Simulator):
        """To be implemented in subclasses."""
        pass

    def read_commands_from_file(self, commands_path: str):
        "Read list of commands from commands.json file"
        with open(commands_path, "r", encoding="utf-8") as f:
            commands = json.load(f)
        return commands

    def simulate(
        self,
        simulator: Simulator,
        output_dir: Optional[Path] = None,
        resource_pool_id: Optional[UUID] = None,
        **kwargs,
    ) -> Path:
        """Simulates the scenario for a single simulator call."""
        output_dir = self._setup_output_dir(output_dir)
        with tempfile.TemporaryDirectory() as input_dir:
            args = self._setup_config(simulator, input_dir)
            return simulator.run(
                input_dir,
                *args,
                output_dir=output_dir,
                resource_pool_id=resource_pool_id,
                **kwargs,
            )

    def simulate_async(
        self,
        simulator: Simulator,
        resource_pool_id: Optional[UUID] = None,
        **kwargs,
    ) -> str:
        """Simulates the scenario asychronously."""

        with tempfile.TemporaryDirectory() as input_dir:
            args = self._setup_config(simulator, input_dir)
            task_id = simulator.run_async(
                input_dir,
                *args,
                resource_pool_id=resource_pool_id,
                **kwargs,
            )
        return task_id



"""Methods to interact with the tasks submitted to the API."""
import pathlib
from typing import Dict, List, Optional, Type

from absl import logging

import inductiva
from inductiva import api
from inductiva.client import ApiClient, ApiException
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.client.model.task_status_code import TaskStatusCode
from inductiva.types import Path
from inductiva.utils.data import unpack_output


def get_task_status(task_id: str):
    """Get the status of a simulation."""

    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        status = api.get_task_status(api_instance, task_id)
    logging.info("Task status: %s", status)

    return status


def fetch_task_output(task_id: str,
                      output_dir: Optional[Path] = None,
                      return_type: Type = pathlib.Path) -> Type:
    """Fetch the results of a simulation."""
    logging.info("Fetching the simulation output for task %s", task_id)
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        status = api.get_task_status(api_instance, task_id)

        logging.info(status)
        if status == "started":
            logging.info("The task is still being executed.")
            return_type = None
        elif status == "submitted":
            logging.info("The task is waiting for resources...")
            return_type = None
        else:
            result_list = api.download_output(api_instance, task_id, output_dir)

            return unpack_output(result_list, output_dir, return_type)


def get_task_info(task_id) -> Dict:
    """Get info about a task by its ID."""
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        try:
            # Get User Tasks
            api_response = api_instance.get_task(
                path_params={"task_id": task_id})

            # Convert DynamicSchema to dict
            return {**api_response.body}

        except ApiException as e:
            raise e


def get_tasks_info(status: Optional[str] = None,
                   page=1,
                   per_page=10) -> List[Dict]:
    """Get information about a user's tasks on the API.

    Tags can be filtered by a status. Results are paginated indexed from 1.
    """
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        query_params = {
            "page": page,
            "per_page": per_page,
        }

        if status is not None:
            query_params["status"] = TaskStatusCode(status)

        try:
            # Get User Tasks
            api_response = api_instance.get_user_tasks(
                query_params=query_params,)

            return [{**task} for task in api_response.body]

        except ApiException as e:
            raise e



"""Methods related to managing executers and resource pools."""
from typing import Optional
from uuid import UUID

import inductiva
from inductiva import api
from inductiva.client import ApiClient, ApiException
from inductiva.client.apis.tags.executers_api import ExecutersApi
from inductiva.client.apis.tags.instance_api import InstanceApi
from inductiva.client.model.instance import Instance
from inductiva.client.model.instance_create import InstanceCreate
from inductiva.client.model.instance_group import InstanceGroup


def create_resource_pool() -> UUID:
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = ExecutersApi(client)

        try:
            api_response = api_instance.create_resource_pool()
        except ApiException as e:
            raise e

    return UUID(api_response.body["id"])


def launch_executer(
    name: str,
    machine_type: str,
    executer_type: str,
    spot: bool = True,
    resource_pool_id: Optional[UUID] = None,
    disk_size_gb: int = 10,
) -> None:

    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)

        body = InstanceCreate(
            name=name,
            machine_type=machine_type,
            image_name=executer_type,
            spot=spot,
            resource_pool_id=resource_pool_id,
            disk_size_gb=disk_size_gb,
        )

        try:
            api_instance.create_instance(body=body)
        except ApiException as e:
            raise e


def kill_executer(name: str) -> None:
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)
        try:
            api_instance.delete_instance(body=Instance(name=name))
        except ApiException as e:
            raise e


def launch_executer_group(
    num_executers: int,
    name: str,
    machine_type: str,
    executer_type: str,
    spot: bool = True,
    resource_pool_id: Optional[UUID] = None,
    disk_size_gb: int = 10,
) -> None:

    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)

        body = InstanceGroup(
            name=name,
            machine_type=machine_type,
            image_name=executer_type,
            spot=spot,
            resource_pool_id=resource_pool_id,
            num_instances=num_executers,
            disk_size_gb=disk_size_gb,
        )

        try:
            api_instance.create_instance_group(body=body)
        except ApiException as e:
            raise e


def kill_executer_group(name: str) -> None:
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = InstanceApi(client)
        try:
            api_instance.delete_instance_group(body=Instance(name=name))
        except ApiException as e:
            raise e



"""Describes the physical scenarios and runs its simulation via API."""
from typing import List, Literal, Optional
from enum import Enum
from dataclasses import dataclass
from uuid import UUID

from inductiva.simulation import Simulator
from inductiva.fluids.simulators import DualSPHysics
from inductiva.fluids.scenarios.fluid_block import FluidBlock
from inductiva.fluids.fluid_types import FluidType
from inductiva.fluids.fluid_types import WATER
from inductiva.types import Path

from inductiva.fluids.scenarios._post_processing import SPHSimulationOutput


@dataclass
class ParticleRadius(Enum):
    """Sets particle radius according to resolution."""
    HIGH = 0.008
    MEDIUM = 0.012
    LOW = 0.02


class DamBreak(FluidBlock):
    """Physical scenario of a dam break simulation."""

    def __init__(
        self,
        fluid: FluidType = WATER,
        dimensions: Optional[List[float]] = None,
        position: Optional[List[float]] = None,
    ):
        """Initializes a `DamBreak` object.

        Args:
            fluid: A fluid type to simulate.
            dimensions: A list containing fluid column dimensions,
              in meters.
            pisition: A list containing fluid column position in a tank,
              in meters.
            """

        if dimensions is None:
            dimensions = [0.3, 1, 1]

        super().__init__(density=fluid.density,
                         kinematic_viscosity=fluid.kinematic_viscosity,
                         dimensions=dimensions,
                         position=position)

    # pylint: disable=arguments-renamed
    def simulate(
        self,
        simulator: Simulator = DualSPHysics(),
        output_dir: Optional[Path] = None,
        resource_pool_id: Optional[UUID] = None,
        device: Literal["cpu", "gpu"] = "gpu",
        resolution: Literal["high", "medium", "low"] = "medium",
        simulation_time: float = 1,
    ):
        """Simulates the scenario.

        Args:
            simulator: Simulator to use.
            output_dir: Directory to store the simulation output.
            device: Device in which to run the simulation.
            resolution: Resolution of the simulation.
            simulation_time: Simulation time, in seconds.
        """

        particle_radius = ParticleRadius[resolution.upper()].value

        sim_output_path = super().simulate(simulator=simulator,
                                           output_dir=output_dir,
                                           resource_pool_id=resource_pool_id,
                                           device=device,
                                           particle_radius=particle_radius,
                                           simulation_time=simulation_time)

        return SPHSimulationOutput(sim_output_path.sim_output_dir)



"""Classes describing fluid tank scenarios."""

from dataclasses import dataclass, field
from enum import Enum
from functools import singledispatchmethod
import os
from typing import List, Literal, Optional
from uuid import UUID

from inductiva.types import Path
from inductiva.scenarios import Scenario
from inductiva.simulation import Simulator
from inductiva.fluids.shapes import BaseShape
from inductiva.fluids.shapes import Rectangle
from inductiva.fluids.shapes import Circle
from inductiva.fluids.shapes import Cube
from inductiva.fluids.shapes import Cylinder
from inductiva.fluids.fluid_types import FluidType
from inductiva.fluids.fluid_types import WATER
from inductiva.fluids.simulators import SPlisHSPlasH
from inductiva.fluids.post_processing.splishsplash import convert_vtk_data_dir_to_netcdf
from inductiva.utils.templates import replace_params_in_template

from . import mesh_file_utils

SPLISHSPLASH_TEMPLATE_FILENAME = "fluid_tank_template.splishsplash.json.jinja"
SPLISHSPLASH_CONFIG_FILENAME = "fluid_tank.json"
TANK_MESH_FILENAME = "tank.obj"
FLUID_MESH_FILENAME = "fluid.obj"


@dataclass
class ParticleRadius(Enum):
    """Sets particle radius according to resolution."""
    HIGH = 0.005
    MEDIUM = 0.01
    LOW = 0.02


@dataclass
class TimeStep(Enum):
    """Sets time step according to resolution."""
    HIGH = 0.0025
    MEDIUM = 0.005
    LOW = 0.01


# Tank inlets.
@dataclass
class BaseTankInlet:
    """Base tank inlet."""
    fluid_velocity: float = 1
    position: List[float] = field(default_factory=lambda: [0, 0])


@dataclass
class RectangularTankInlet(BaseTankInlet, Rectangle):
    """Rectangular tank inlet."""
    pass


@dataclass
class CircularTankInlet(BaseTankInlet, Circle):
    """Circular tank inlet."""
    pass


# Tank outlets.
class BaseTankOutlet:
    """Base tank outlet."""
    pass


class CubicTankOutlet(BaseTankOutlet):
    """Cubic tank outlet."""

    def __init__(
        self,
        dimensions: List[float],
        top_base_position: Optional[List[float]] = None,
    ):
        """Initializes a cubic tank outlet.

        Args:
            dimensions: Dimensions of the outlet.
            top_base_position: Position of the top base of the outlet.
        """
        top_base_position = top_base_position or [0, 0]

        self.shape = Cube(
            dimensions=dimensions,
            position=[*top_base_position, -dimensions[2]],
        )


class CylindricalTankOutlet(BaseTankOutlet):
    """Cylindrical tank outlet."""

    def __init__(
        self,
        radius: float,
        height: float,
        top_base_position: Optional[List[float]] = None,
    ):
        """Initializes a cylindrical tank outlet.

        Args:
            radius: Radius of the outlet.
            height: Height of the outlet.
            top_base_position: Position of the top base of the outlet.
        """
        top_base_position = top_base_position or [0, 0]

        self.shape = Cylinder(
            radius=radius,
            height=height,
            position=[*top_base_position, -height],
        )


class FluidTank(Scenario):
    """Fluid tank."""

    def __init__(
        self,
        shape: BaseShape = Cylinder(radius=0.5, height=1),
        fluid: FluidType = WATER,
        fluid_level: float = 0,
        inlet: Optional[BaseTankInlet] = CircularTankInlet(radius=0.1),
        outlet: Optional[BaseTankOutlet] = CylindricalTankOutlet(radius=0.1,
                                                                 height=0.1),
    ):
        """Initializes a fluid tank.

        Args:
            shape: Shape of the tank.
            fluid: Fluid type.
            fluid_level: Fluid level initially in the tank.
            inlet: Inlet of the tank.
            outlet: Outlet of the tank.
        """
        self.shape = shape
        self.fluid = fluid
        self.fluid_level = fluid_level
        self.inlet = inlet
        self.outlet = outlet

    def simulate(
        self,
        simulator: Simulator = SPlisHSPlasH(),
        output_dir: Optional[Path] = None,
        resource_pool_id: Optional[UUID] = None,
        device: Literal["cpu", "gpu"] = "cpu",
        simulation_time: float = 5,
        resolution: Literal["low", "medium", "high"] = "low",
        output_time_step: float = 0.1,
        particle_sorting: bool = False,
    ):
        """Simulates the scenario.

        Args:
            simulator: Simulator to use.
            output_dir: Directory to store the simulation output.
            device: Device in which to run the simulation.
            simulation_time: Simulation time, in seconds.
            output_time_step: Time step between output files, in seconds.
            resolution: Resolution of the simulation. Controls the particle
                radius and time step.
            particle_sorting: Whether to use particle sorting.
        """

        self.simulation_time = simulation_time
        self.particle_radius = ParticleRadius[resolution.upper()].value
        self.time_step = TimeStep[resolution.upper()].value
        self.output_time_step = output_time_step
        self.particle_sorting = particle_sorting

        output_path = super().simulate(
            simulator,
            output_dir=output_dir,
            resource_pool_id=resource_pool_id,
            device=device,
        )

        # TODO: Add any kind of post-processing here, e.g. convert files?
        convert_vtk_data_dir_to_netcdf(
            data_dir=os.path.join(output_path, "vtk"),
            output_time_step=self.output_time_step,
            netcdf_data_dir=os.path.join(output_path, "netcdf"))

        return output_path

    def get_bounding_box(self):
        """Gets the bounding box of the tank.

        Returns:
            Tuple of two lists representing the minimum and maximum coordinates
            of the bounding box of the tank, respectively.
        """

        bounding_box_min, bounding_box_max = self.shape.get_bounding_box()

        # Extend the bounding box to include the outlet.
        if self.outlet is not None:
            outlet_bounding_box_min, _ = self.outlet.shape.get_bounding_box()
            bounding_box_min[2] = outlet_bounding_box_min[2]

        return bounding_box_min, bounding_box_max

    @singledispatchmethod
    @classmethod
    def get_config_filename(cls, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{cls.__name__}` scenario.")

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_config(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@FluidTank.get_config_filename.register
def _(cls, simulator: SPlisHSPlasH) -> str:  # pylint: disable=unused-argument
    """Returns the config filename for SPlisHSPlasH."""
    return SPLISHSPLASH_CONFIG_FILENAME


@FluidTank.gen_aux_files.register
def _(self, simulator: SPlisHSPlasH, input_dir):  # pylint: disable=unused-argument
    """Generates auxiliary files for SPlisHSPlasH."""
    mesh_file_utils.create_tank_mesh_file(
        shape=self.shape,
        outlet=self.outlet,
        path=os.path.join(input_dir, TANK_MESH_FILENAME),
    )

    mesh_file_utils.create_tank_fluid_mesh_file(
        shape=self.shape,
        fluid_level=self.fluid_level,
        margin=2 * self.particle_radius,
        path=os.path.join(input_dir, FLUID_MESH_FILENAME),
    )


@FluidTank.gen_config.register
def _(self, simulator: SPlisHSPlasH, input_dir: str):  # pylint: disable=unused-argument
    """Generates the configuration file for SPlisHSPlasH."""

    bounding_box_min, bounding_box_max = self.get_bounding_box()
    inlet_position = [
        self.inlet.position[0],
        self.inlet.position[1],
        bounding_box_max[2],
    ]

    replace_params_in_template(
        templates_dir=os.path.dirname(__file__),
        template_filename=SPLISHSPLASH_TEMPLATE_FILENAME,
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_radius": self.particle_radius,
            "data_export_rate": 1 / self.output_time_step,
            "z_sort": self.particle_sorting,
            "tank_filename": TANK_MESH_FILENAME,
            "fluid_filename": FLUID_MESH_FILENAME,
            "fluid": self.fluid,
            "inlet_position": inlet_position,
            "inlet_width": int(self.inlet.radius / self.particle_radius),
            "inlet_fluid_velocity": self.inlet.fluid_velocity,
            "bounding_box_min": bounding_box_min,
            "bounding_box_max": bounding_box_max,
        },
        output_file_path=os.path.join(input_dir, SPLISHSPLASH_CONFIG_FILENAME),
    )



# pylint: disable=missing-module-docstring
import re


def split_camel_case(s):
    """Split a camel case string into a list of strings."""
    return re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1",
                                                  s)).split()



"""Molecular Dynamics simulation for water box scenario."""
from functools import singledispatchmethod
from typing import Optional, Literal
import os
import shutil
from uuid import UUID

from inductiva.types import Path
from inductiva.molecules.simulators import GROMACS
from inductiva.simulation import Simulator
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       batch_replace_params_in_template,
                                       replace_params_in_template)
from inductiva.scenarios import Scenario
from inductiva.utils.files import remove_files_with_tag

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "md_water_box")
GROMACS_TEMPLATE_INPUT_DIR = "gromacs"


class MDWaterBox(Scenario):
    """Molecular dynamics water box scenario."""

    def __init__(
        self,
        temperature: float = 300,
        box_size: float = 2.3,
    ):
        """The scenario involves simulating a box with water molecules.
        The simulation consists of two main steps: energy minimization
        and molecular dynamics simulation.
        By default, the initial positions of the water molecules are
        arranged uniformally, so to randomize them we perform a
        decorrelation step.
        Args:
            temperature: The temperature of the simulation in Kelvin.
            box_size: The size of the box in nm.
        """
        self.template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                         GROMACS_TEMPLATE_INPUT_DIR)
        self.temperature = temperature
        if box_size < 2.3:
            raise ValueError("The box size must be greater than 2.3 nm.")
        self.box_size = box_size

    def simulate(
            self,
            simulator: Simulator = GROMACS(),
            output_dir: Optional[Path] = None,
            resource_pool_id: Optional[UUID] = None,
            simulation_time: float = 10,  # ns
            integrator: Literal["md", "sd", "bd"] = "md",
            nsteps_minim: int = 5000):
        """Simulate the water box scenario using molecular dynamics.

        Args:
            output_dir: The output directory to save the simulation results.
            simulation_time: The simulation time in ns.
            integrator: The integrator to use for the simulation. Options:
                - "md" (Molecular Dynamics): Accurate leap-frog algorithm for
                integrating Newton's equations of motion.
                - "sd" (Steepest Descent): Stochastic dynamics integrator with
                leap-frog scheme.
                - "bd" (Brownian Dynamics): Euler integrator for Brownian or
                position Langevin dynamics.

            For more details on the integrators, refer to the GROMACS
            documentation at
            https://manual.gromacs.org/current/user-guide/mdp-options.html.

            nsteps_minim: Number of steps for energy minimization.
        """
        self.nsteps = int(
            simulation_time * 1e6 / 2
        )  # convert to fs and divide by the time step of the simulation (2 fs)
        self.integrator = integrator
        self.nsteps_minim = nsteps_minim
        commands_path = os.path.join(self.template_dir, "commands.json")
        replace_params_in_template(self.template_dir, "commands.json.jinja",
                                   {"box_size": self.box_size}, commands_path)
        commands = self.read_commands_from_file(commands_path)
        return super().simulate(simulator,
                                output_dir,
                                resource_pool_id=resource_pool_id,
                                commands=commands)

    def simulate_async(
            self,
            simulator: Simulator = GROMACS(),
            resource_pool_id: Optional[UUID] = None,
            simulation_time: float = 10,  # ns
            integrator: Literal["md", "sd", "bd"] = "md",
            nsteps_minim: int = 5000):
        """Simulate the water box scenario using molecular dynamics
        asyncronously.

        Args:
            simulation_time: The simulation time in ns.
            integrator: The integrator to use for the simulation. Options:
                - "md" (Molecular Dynamics): Accurate leap-frog algorithm for
                integrating Newton's equations of motion.
                - "sd" (Steepest Descent): Stochastic dynamics integrator with
                leap-frog scheme.
                - "bd" (Brownian Dynamics): Euler integrator for Brownian or
                position Langevin dynamics.

            For more details on the integrators, refer to the GROMACS
            documentation at
            https://manual.gromacs.org/current/user-guide/mdp-options.html.

            nsteps_minim: Number of steps for energy minimization.
        """
        self.nsteps = int(
            simulation_time * 1e6 / 2
        )  # convert to fs and divide by the time step of the simulation (2 fs)
        self.integrator = integrator
        self.nsteps_minim = nsteps_minim
        commands_path = os.path.join(self.template_dir, "commands.json")
        replace_params_in_template(self.template_dir, "commands.json.jinja",
                                   {"box_size": self.box_size}, commands_path)
        commands = self.read_commands_from_file(commands_path)
        return super().simulate_async(simulator,
                                      resource_pool_id=resource_pool_id,
                                      commands=commands)

    @singledispatchmethod
    def gen_config(self, simulator: Simulator):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def get_config_filename(self, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@MDWaterBox.get_config_filename.register
def _(self, simulator: GROMACS):  # pylint: disable=unused-argument
    pass


@MDWaterBox.gen_aux_files.register
def _(self, simulator: GROMACS, input_dir):  # pylint: disable=unused-argument
    """Setup the working directory for the simulation."""
    shutil.copytree(os.path.join(self.template_dir),
                    os.path.join(input_dir),
                    dirs_exist_ok=True)
    remove_files_with_tag(input_dir, ".jinja")


@MDWaterBox.gen_config.register
def _(self, simulator: GROMACS, input_dir):  # pylint: disable=unused-argument
    """Generate the mdp configuration files for the simulation."""
    batch_replace_params_in_template(
        templates_dir=self.template_dir,
        template_filename_paths=[
            "simulation.mdp.jinja",
            "energy_minimization.mdp.jinja",
        ],
        params={
            "integrator": self.integrator,
            "nsteps": self.nsteps,
            "ref_temp": self.temperature,
            "nsteps_minim": self.nsteps_minim,
        },
        output_filename_paths=[
            os.path.join(input_dir, "simulation.mdp"),
            os.path.join(input_dir, "energy_minimization.mdp"),
        ])



"""Heat sink scenario."""
from functools import singledispatchmethod
import os
import shutil
from typing import Optional
from uuid import UUID

from inductiva.types import Path
from inductiva.fluids.simulators import OpenFOAM
from inductiva.simulation import Simulator
from inductiva.scenarios import Scenario
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       replace_params_in_template)

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "heat_sink")
OPENFOAM_TEMPLATE_SUBDIR = "openfoam"
FILES_SUBDIR = "files"
OPENFOAM_TEMPLATE_PARAMS_FILE_NAME = "parameters.jinja"
OPENFOAM_PARAMS_FILE_NAME = "parameters"
COMMANDS_FILE_NAME = "commands.json"


class HeatSink(Scenario):
    """Heat sink scenario.
    
    This is a simulation scenario for a heat sink. A heat source is placed
    in a box where there is an air flow. A heat sink, placed on top of the 
    source, is used to dissipate the heat via convection with the air flow.

    The heat source is modeled as a heater with a given power. The heat sink
    is a block of aluminum with thin fins on top. The air flow is introduced
    in the simulation via an inlet, where the air is at a fixed temperature.
    """

    def __init__(
        self,
        air_velocity=10,
        air_temperature=290,
        heater_power=200,
    ):
        """Initializes the heat sink scenario.
        Args:
            air_velocity: The velocity of the air flow, in m/s.
            air_temperature: The temperature of the air flow, in Kelvin. Also
              sets the initial temperature of the heater and the heat sink.
            heater_power: The power of the heater, in Watts.
        """
        self.air_velocity = air_velocity
        self.air_temperature = air_temperature
        self.heater_power = heater_power

    def simulate(
        self,
        simulator: Simulator = OpenFOAM(),
        output_dir: Optional[Path] = None,
        resource_pool_id: Optional[UUID] = None,
        simulation_time=300,
        output_time_step=10,
    ):
        """Simulates the scenario.

        Args:
            simulator: The simulator to use for the simulation.
            output_dir: The output directory to save the simulation results.
            simulation_time: The simulation time, in seconds.
            output_time_step: The time step to save the simulation results, in
              seconds.
        """
        self.simulation_time = simulation_time
        self.output_time_step = output_time_step

        commands = self.get_commands()

        return super().simulate(simulator,
                                output_dir,
                                resource_pool_id=resource_pool_id,
                                commands=commands)

    def simulate_async(
        self,
        simulator: Simulator = OpenFOAM(),
        resource_pool_id: Optional[UUID] = None,
        simulation_time=300,
        output_time_step=10,
    ):
        """Simulates the scenario asynchronously.

        Args:
            simulator: The simulator to use for the simulation.
            output_dir: The output directory to save the simulation results.
            simulation_time: The simulation time, in seconds.
            output_time_step: The time step to save the simulation results, in
              seconds.
        """
        self.simulation_time = simulation_time
        self.output_time_step = output_time_step

        commands = self.get_commands()

        return super().simulate_async(simulator,
                                      resource_pool_id=resource_pool_id,
                                      commands=commands)

    def get_commands(self):
        """Returns the commands for the simulation.
        """
        commands_file_path = os.path.join(SCENARIO_TEMPLATE_DIR,
                                          OPENFOAM_TEMPLATE_SUBDIR,
                                          COMMANDS_FILE_NAME)

        commands = self.read_commands_from_file(commands_file_path)

        return commands

    @singledispatchmethod
    def gen_config(self, simulator: Simulator):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def get_config_filename(self, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@HeatSink.get_config_filename.register
def _(self, simulator: OpenFOAM):  # pylint: disable=unused-argument
    pass


@HeatSink.gen_aux_files.register
def _(self, simulator: OpenFOAM, input_dir):  # pylint: disable=unused-argument
    """Setup the working directory for the simulation."""

    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      OPENFOAM_TEMPLATE_SUBDIR, FILES_SUBDIR)

    shutil.copytree(template_files_dir,
                    input_dir,
                    dirs_exist_ok=True,
                    symlinks=True)

    params_file_path = os.path.join(input_dir, OPENFOAM_PARAMS_FILE_NAME)

    replace_params_in_template(
        templates_dir=input_dir,
        template_filename=OPENFOAM_TEMPLATE_PARAMS_FILE_NAME,
        params={
            "simulation_time": self.simulation_time,
            "output_time_step": self.output_time_step,
            "air_velocity": self.air_velocity,
            "air_temperature": self.air_temperature,
            "heater_power": self.heater_power
        },
        output_file_path=params_file_path,
    )

    # TODO (fabiocruz): add option to remove template file in
    # replace_params_in_template.
    os.remove(os.path.join(input_dir, OPENFOAM_TEMPLATE_PARAMS_FILE_NAME))


@HeatSink.gen_config.register
def _(self, simulator: OpenFOAM, input_dir):  # pylint: disable=unused-argument
    """Generate the mdp configuration files for the simulation."""
    pass



"""Post process Gromacs simulation outputs."""
import glob
import os
import MDAnalysis as mda
import nglview as nv

from inductiva.types import Path
from inductiva.utils.templates import (TEMPLATES_PATH)

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "protein_visualization")
GROMACS_TEMPLATE_INPUT_DIR = "gromacs"


class GROMACSSimulationOutput:
    """Post process a GROMACS simulation outputs."""

    def __init__(self, sim_output_path: Path = None):
        """Initializes a `SimulationOutput` object.

        Given a simulation output directory that contains the standard files
        generated by a GROMACS simulation run, this class provides methods to
        visualize the simulation outputs in a notebook.

        Args:
            sim_output_path: Path to the simulation output directory."""

        self.sim_output_dir = sim_output_path

    def render(self,
               pdb_file_name: str = None,
               trajectory_file_name: str = "trajectory.xtc"):
        """Visualize the simulation outputs in a notebook using NGLView.

        Args:
            pdb_file: Path to the PDB file to be visualized.
            trajectory_name: Name of the trajectory file to be visualized."""

        if pdb_file_name is None:
            pdb_pattern = os.path.join(self.sim_output_dir, "*.pdb")
            pdb_file_name = glob.glob(pdb_pattern)

            if pdb_file_name is None:
                raise ValueError("No PDB file found in the output directory.")

            if len(pdb_file_name) != 1:
                raise ValueError(
                    "Please specify the .pdb file to be visualized.")

        protein_file = os.path.join(self.sim_output_dir, pdb_file_name)
        trajectory = os.path.join(self.sim_output_dir, trajectory_file_name)
        system = mda.Universe(protein_file, trajectory)

        view = nv.show_mdanalysis(system)
        view.add_ball_and_stick(
            "all")  # Render the molecules as ball-and-stick models
        view.center()  # Center the view
        view.parameters = {
            "backgroundColor": "white"
        }  # Set the background color

        return view



#pylint: disable=missing-module-docstring
from .dam_break import DamBreak
from .fluid_block import FluidBlock
from .wind_tunnel import WindTunnel
from .wind_tunnel import WindTunnelSimulationOutput
from .fluid_tank import (
    FluidTank,
    CubicTankOutlet,
    CylindricalTankOutlet,
    RectangularTankInlet,
    CircularTankInlet,
)
from .heat_sink import HeatSink



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.admin_users.post import AddUser
from inductiva.client.paths.admin_tasks.get import GetTasks
from inductiva.client.paths.admin_users_username_tasks.get import GetTasksByUsername
from inductiva.client.paths.admin_users_username.get import GetUser


class AdminApi(
        AddUser,
        GetTasks,
        GetTasksByUsername,
        GetUser,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class Executer(
        schemas.ComposedSchema,):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Union of all executer types.
    """

    class MetaOapg:

        @staticmethod
        def discriminator():
            return {
                'host_type': {
                    'GCloudExecuter': GCloudExecuter,
                    'InductivaExecuter': InductivaExecuter,
                    'gcloud': GCloudExecuter,
                    'inductiva-hardware': InductivaExecuter,
                }
            }

        @classmethod
        @functools.lru_cache()
        def one_of(cls):
            # we need this here to make our import statements work
            # we must store _composed_schemas in here so the code is only run
            # when we invoke this method. If we kept this at the class
            # level we would get an error because the class level
            # code would be run when this module is imported, and these composed
            # classes don't exist yet because their module has not finished
            # loading
            return [
                GCloudExecuter,
                InductivaExecuter,
            ]

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
            str,
            date,
            datetime,
            uuid.UUID,
            int,
            float,
            decimal.Decimal,
            bool,
            None,
            list,
            tuple,
            bytes,
            io.FileIO,
            io.BufferedReader,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'Executer':
        return super().__new__(
            cls,
            *_args,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.g_cloud_executer import GCloudExecuter
from inductiva.client.model.inductiva_executer import InductivaExecuter



#pylint: disable=missing-module-docstring
from .splishsplash import SPlisHSPlasH
from .dualsphysics import DualSPHysics
from .xbeach import XBeach
from .swash import SWASH
from .openfoam import OpenFOAM



import typing_extensions

from inductiva.client.apis.tags import TagValues
from inductiva.client.apis.tags.admin_api import AdminApi
from inductiva.client.apis.tags.executers_api import ExecutersApi
from inductiva.client.apis.tags.instance_api import InstanceApi
from inductiva.client.apis.tags.tasks_api import TasksApi

TagToApi = typing_extensions.TypedDict(
    'TagToApi', {
        TagValues.ADMIN: AdminApi,
        TagValues.EXECUTERS: ExecutersApi,
        TagValues.INSTANCE: InstanceApi,
        TagValues.TASKS: TasksApi,
    })

tag_to_api = TagToApi({
    TagValues.ADMIN: AdminApi,
    TagValues.EXECUTERS: ExecutersApi,
    TagValues.INSTANCE: InstanceApi,
    TagValues.TASKS: TasksApi,
})



"""DualSPHysics module of the API."""
import pathlib
from typing import Literal, Optional
from uuid import UUID

from inductiva import types
from inductiva.simulation import Simulator


class SPlisHSPlasH(Simulator):
    """Class to invoke a generic SPlisHSPlasH simulation on the API."""

    @property
    def api_method_name(self) -> str:
        return "sph.splishsplash.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        output_dir: Optional[types.Path] = None,
        resource_pool_id: Optional[UUID] = None,
        device: Literal["gpu", "cpu"] = "cpu",
    ) -> pathlib.Path:
        """Run the simulation.

        Args:
            sim_config_filename: Name of the simulation configuration file.
            device: Device in which to run the simulation.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            output_dir=output_dir,
            resource_pool_id=resource_pool_id,
            device=device,
            input_filename=sim_config_filename,
        )

    def run_async(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        resource_pool_id: Optional[UUID] = None,
        device: Literal["gpu", "cpu"] = "cpu",
    ) -> str:
        """Run the simulation asynchronously.

        Args:
            sim_config_filename: Name of the simulation configuration file.
            device: Device in which to run the simulation.
            other arguments: See the documentation of the base class.
            """

        return super().run_async(input_dir,
                                 resource_pool_id=resource_pool_id,
                                 device=device,
                                 input_filename=sim_config_filename)



"""Utils related to metaprogramming."""
import inspect

# pylint: disable=protected-access


def is_tuple(type_annotation) -> bool:
    """Check if a type annotation is a tuple.

    Note: the current implementation of this function may not work for all
    python versions.

    Args:
        type_annotation: Type annotation to check.

    Returns:
        Bool that reflects if a type annotation is a tuple.
    """
    # Some of the Python versions tested have the attribute _name while others
    # have __name__
    if hasattr(type_annotation, "_name"):
        return type_annotation._name == "Tuple"
    if hasattr(type_annotation, "__name__"):
        return type_annotation.__name__ == "Tuple"

    return False


# pylint: enable=protected-access


def get_type_annotations(function_ptr) -> dict:
    """Get type annotations of a given function.

    Args:
        function_ptr: Function to analyse.

    Return:
        Returns a dict with type annotations. "return" key specifies
        the type annotation of the return.
    """
    return inspect.getfullargspec(function_ptr).annotations


def get_method_name(function_ptr) -> str:
    """Constructs the name of a method supported by the Web API.

    Example: If `function_ptr` is the function `sum` placed in the
    `inductiva.core.math` module, it returns the string `math.sum`.

    Args:
        function_ptr: Function to analyse.

    Return:
        Returns name of the method.
    """
    module_name = function_ptr.__module__
    module_name_useful_segments = module_name.split(".")[2:]
    module_name = ".".join(module_name_useful_segments)

    return f"{module_name}.{function_ptr.__name__}"



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from inductiva.client.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from inductiva.client.model.body_upload_task_input import BodyUploadTaskInput
from inductiva.client.model.executer import Executer
from inductiva.client.model.executer_create import ExecuterCreate
from inductiva.client.model.executer_id import ExecuterID
from inductiva.client.model.g_cloud_executer import GCloudExecuter
from inductiva.client.model.g_cloud_host_info import GCloudHostInfo
from inductiva.client.model.http_validation_error import HTTPValidationError
from inductiva.client.model.inductiva_executer import InductivaExecuter
from inductiva.client.model.inductiva_host_info import InductivaHostInfo
from inductiva.client.model.instance import Instance
from inductiva.client.model.instance_create import InstanceCreate
from inductiva.client.model.instance_group import InstanceGroup
from inductiva.client.model.resource_pool_id import ResourcePoolID
from inductiva.client.model.task import Task
from inductiva.client.model.task_request import TaskRequest
from inductiva.client.model.task_status import TaskStatus
from inductiva.client.model.task_status_code import TaskStatusCode
from inductiva.client.model.task_with_user_info import TaskWithUserInfo
from inductiva.client.model.user import User
from inductiva.client.model.user_create import UserCreate
from inductiva.client.model.username import Username
from inductiva.client.model.validation_error import ValidationError



"""Protein solvation scenario."""

from functools import singledispatchmethod
from typing import Optional, Literal
import os
import shutil
from absl import logging
import numpy as np
from uuid import UUID

from inductiva.types import Path
from inductiva.molecules.simulators import GROMACS
from inductiva.simulation import Simulator
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       batch_replace_params_in_template,
                                       replace_params_in_template)
from inductiva.scenarios import Scenario
from inductiva.utils.files import remove_files_with_tag
from inductiva.utils import files

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "protein_solvation")
GROMACS_TEMPLATE_INPUT_DIR = "gromacs"


class ProteinSolvation(Scenario):
    """Solvated protein scenario."""

    def __init__(self,
                 protein_pdb: str,
                 temperature: float = 300,
                 charged: bool = None):
        """
        Scenario constructor for protein solvation based on the GROMACS
        simulator.
        The three main steps of this scenario are solvation, energy minimization
        and simulation. The user can control the number of steps used to perform
        the energy minimization step and the duration, temperature and
        integrator used to perform the simulation.
        Args:
            protein_pdb: The path to the protein pdb file.
            temperature: The temperature to use for the simulation.
            charged: Whether the protein is charged or not. If None, the charge
            is computed automatically.
        """
        self.template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                         GROMACS_TEMPLATE_INPUT_DIR)
        self.protein_pdb = files.resolve_path(protein_pdb)
        self.temperature = temperature
        self.charged = charged

    def simulate(
            self,
            simulator: Simulator = GROMACS(),
            output_dir: Optional[Path] = None,
            resource_pool_id: Optional[UUID] = None,
            simulation_time: float = 10,  # ns
            integrator: Literal["md", "sd", "bd"] = "md",
            nsteps_minim: int = 5000,
            visualized_section: str = "Protein-H"):
        """Simulate the solvation of a protein.

        Args:
            output_dir: The output directory to save the simulation results.
            simulation_time: The simulation time in ns.
            integrator: The integrator to use for the simulation. Options:
                - "md" (Molecular Dynamics): Accurate leap-frog algorithm for
                integrating Newton's equations of motion.
                - "sd" (Steepest Descent): Stochastic dynamics integrator with
                leap-frog scheme.
                - "bd" (Brownian Dynamics): Euler integrator for Brownian or
                position Langevin dynamics.

            For more details on the integrators, refer to the GROMACS
            documentation at
            https://manual.gromacs.org/current/user-guide/mdp-options.html.

            nsteps_minim: Number of steps for energy minimization.
            visualized_section: The section of the protein to visualize in the
            simulation.
                Options:
                - "Protein-H": The protein with hydrogens. This is the default.
                - "System": The whole system (Protein + Water).
        """

        #Compute charge if it is not provided
        if self.charged is None:
            self.charged = self.compute_charge(
                resource_pool_id=resource_pool_id)

        #Edit commands.json according to the charge of the protein
        commands_path = os.path.join(self.template_dir, "commands.json")

        replace_params_in_template(
            self.template_dir, "commands.json.jinja", {
                "pdb_file": self.protein_pdb,
                "charged": self.charged,
                "visualized_section": visualized_section
            }, commands_path)

        commands = self.read_commands_from_file(commands_path)
        self.nsteps = int(
            simulation_time * 1e6 / 2
        )  # convert to fs and divide by the time step of the simulation (2 fs)
        self.integrator = integrator
        self.nsteps_minim = nsteps_minim
        return super().simulate(simulator,
                                output_dir,
                                resource_pool_id=resource_pool_id,
                                commands=commands)

    def simulate_async(
            self,
            simulator: Simulator = GROMACS(),
            resource_pool_id: Optional[UUID] = None,
            simulation_time: float = 10,  # ns
            integrator: Literal["md", "sd", "bd"] = "md",
            nsteps_minim: int = 5000,
            visualized_section: str = "Protein-H"):
        """Simulate the solvation of a protein scenario asyncronously.

        Args:
            simulation_time: The simulation time in ns.
            integrator: The integrator to use for the simulation. Options:
                - "md" (Molecular Dynamics): Accurate leap-frog algorithm for
                integrating Newton's equations of motion.
                - "sd" (Steepest Descent): Stochastic dynamics integrator with
                leap-frog scheme.
                - "bd" (Brownian Dynamics): Euler integrator for Brownian or
                position Langevin dynamics.

            For more details on the integrators, refer to the GROMACS
            documentation at
            https://manual.gromacs.org/current/user-guide/mdp-options.html.

            nsteps_minim: Number of steps for energy minimization.
            visualized_section: The section of the protein to visualize in the
            simulation.
                Options:
                - "Protein-H": The protein with hydrogens. This is the default.
                - "System": The whole system (Protein + Water).
        """

        #Compute charge if it is not provided
        if self.charged is None:
            self.charged = self.compute_charge(
                resource_pool_id=resource_pool_id)

        #Edit commands.json according to the charge of the protein
        commands_path = os.path.join(self.template_dir, "commands.json")

        replace_params_in_template(
            self.template_dir, "commands.json.jinja", {
                "pdb_file": self.protein_pdb,
                "charged": self.charged,
                "visualized_section": visualized_section
            }, commands_path)

        commands = self.read_commands_from_file(commands_path)

        self.nsteps = int(
            simulation_time * 1e6 / 2
        )  # convert to fs and divide by the time step of the simulation (2 fs)
        self.integrator = integrator
        self.nsteps_minim = nsteps_minim

        self.task_id = super().simulate_async(simulator,
                                              resource_pool_id=resource_pool_id,
                                              commands=commands)

        return self.task_id

    def compute_charge(self,
                       simulator: Simulator = GROMACS(),
                       resource_pool_id: Optional[UUID] = None):
        """Check if the protein is charged.

        If the protein is not charged, the charge is computed by reading
        the topology file generated by the GROMACS tool gmx pdb2gmx."""

        logging.info("Computing the charge of the protein")

        protein_directory = os.path.dirname(self.protein_pdb)

        commands_path = os.path.join(self.template_dir,
                                     "charge_computation.json")
        protein_pdb = os.path.basename(self.protein_pdb)
        replace_params_in_template(self.template_dir,
                                   "charge_computation.json.jinja",
                                   {"pdb_file": protein_pdb}, commands_path)

        commands = self.read_commands_from_file(commands_path)

        simulator.run(protein_directory,
                      commands=commands,
                      output_dir=protein_directory,
                      resource_pool_id=resource_pool_id)

        topology_file = os.path.join(protein_directory, "topol.top")

        #Information about the charge of each residue is stored in the topology
        #file in the lines starting with "; residue". Summing the charges of
        #all residues gives the total charge of the protein.

        with open(topology_file, "r", encoding="utf-8") as file:
            charge = np.sum([
                float(line.split()[-1])
                for line in file
                if line.startswith("; residue")
            ])

        is_charged = abs(charge) > 1e-6
        return is_charged

    @singledispatchmethod
    def gen_config(self, simulator: Simulator):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def get_config_filename(self, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@ProteinSolvation.get_config_filename.register
def _(self, simulator: GROMACS):  # pylint: disable=unused-argument
    pass


@ProteinSolvation.gen_aux_files.register
def _(self, simulator: GROMACS, input_dir):  # pylint: disable=unused-argument
    """Setup the working directory for the simulation."""
    # rename the pdb file to comply with the naming in the commands list
    shutil.copy(self.protein_pdb, os.path.join(input_dir, "protein.pdb"))
    shutil.copytree(os.path.join(self.template_dir),
                    os.path.join(input_dir),
                    dirs_exist_ok=True)
    remove_files_with_tag(input_dir, ".jinja")


@ProteinSolvation.gen_config.register
def _(self, simulator: GROMACS, input_dir):  # pylint: disable=unused-argument
    """Generate the mdp configuration files for the simulation."""
    batch_replace_params_in_template(
        templates_dir=self.template_dir,
        template_filename_paths=[
            "simulation.mdp.jinja",
            "energy_minimization.mdp.jinja",
        ],
        params={
            "integrator": self.integrator,
            "nsteps": self.nsteps,
            "ref_temp": self.temperature,
            "nsteps_minim": self.nsteps_minim,
        },
        output_filename_paths=[
            os.path.join(input_dir, "simulation.mdp"),
            os.path.join(input_dir, "energy_minimization.mdp"),
        ])



"""Utilities to parametrically generate meshes with gmsh.

TODO: Move this elsewhere in this repo to be reused by other high-level
scenarios? Or, further in the future, to API endpoints?
"""

import contextlib
import os

import gmsh
import meshio

TMP_MSH_FILE_PATH = "gmsh_generated.msh"
MESH_SIZE = 0.05
GMSH_LOG_LEVEL = 2


class gmshAPIWrapper(contextlib.AbstractContextManager):  # pylint: disable=invalid-name
    """Wrapper to interact with the gmsh API.

    Applies common operations with the gmsh API before and after use
    case-specific primitives.
    """

    def __init__(self,
                 mesh_size: float = MESH_SIZE,
                 msh_file_path: str = TMP_MSH_FILE_PATH,
                 log_level: int = GMSH_LOG_LEVEL):

        self.mesh_size = mesh_size
        self.msh_file_path = msh_file_path
        self.log_level = log_level

    def __enter__(self):
        """Common operations applied before use case-specific primitives."""

        # Initialize the gmsh API.
        gmsh.initialize()

        # Set the log level.
        gmsh.option.setNumber("General.Verbosity", self.log_level)

        # Set the mesh size.
        gmsh.option.setNumber("Mesh.MeshSizeMax", self.mesh_size)

    def __exit__(self, *_):
        """Common operations applied after use case-specific primitives."""

        # Synchronize the representations (or kernels) available in gmsh:
        # - the built-in representation, used in a bottom-up manner (defining
        #   points first, then curves, surfaces and volumes).
        # - the OpenCascade representation, used as the built-in representation
        #   or in a top-down constructive solid geometry fashion (defining
        #   solids on which boolean operations are performed).
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()

        # Generate the mesh elements up to dimension 2, i.e. points, curves and
        # surfaces.
        gmsh.model.mesh.generate(2)

        # Write mesh to a file.
        gmsh.write(self.msh_file_path)

        # Close the gmsh API.
        gmsh.finalize()


def add_x_rectangle_loop(x, y, z, ly, lz):
    p1 = gmsh.model.geo.addPoint(x, y, z, 0)
    p2 = gmsh.model.geo.addPoint(x, y + ly, z, 0)
    p3 = gmsh.model.geo.addPoint(x, y + ly, z + lz, 0)
    p4 = gmsh.model.geo.addPoint(x, y, z + lz, 0)

    c1 = gmsh.model.geo.addLine(p1, p2)
    c2 = gmsh.model.geo.addLine(p2, p3)
    c3 = gmsh.model.geo.addLine(p3, p4)
    c4 = gmsh.model.geo.addLine(p4, p1)

    l = gmsh.model.geo.addCurveLoop([c1, c2, c3, c4])

    return [p1, p2, p3, p4], [c1, c2, c3, c4], l


def add_y_rectangle_loop(x, y, z, lx, lz):
    p1 = gmsh.model.geo.addPoint(x, y, z, 0)
    p2 = gmsh.model.geo.addPoint(x + lx, y, z, 0)
    p3 = gmsh.model.geo.addPoint(x + lx, y, z + lz, 0)
    p4 = gmsh.model.geo.addPoint(x, y, z + lz, 0)

    c1 = gmsh.model.geo.addLine(p1, p2)
    c2 = gmsh.model.geo.addLine(p2, p3)
    c3 = gmsh.model.geo.addLine(p3, p4)
    c4 = gmsh.model.geo.addLine(p4, p1)

    l = gmsh.model.geo.addCurveLoop([c1, c2, c3, c4])

    return [p1, p2, p3, p4], [c1, c2, c3, c4], l


def add_z_rectangle_loop(x, y, z, lx, ly):
    p1 = gmsh.model.geo.addPoint(x, y, z, 0)
    p2 = gmsh.model.geo.addPoint(x + lx, y, z, 0)
    p3 = gmsh.model.geo.addPoint(x + lx, y + ly, z, 0)
    p4 = gmsh.model.geo.addPoint(x, y + ly, z, 0)

    c1 = gmsh.model.geo.addLine(p1, p2)
    c2 = gmsh.model.geo.addLine(p2, p3)
    c3 = gmsh.model.geo.addLine(p3, p4)
    c4 = gmsh.model.geo.addLine(p4, p1)

    l = gmsh.model.geo.addCurveLoop([c1, c2, c3, c4])

    return [p1, p2, p3, p4], [c1, c2, c3, c4], l


def add_x_rectangle(x, y, z, ly, lz, hole_loops):

    # Add the corresponding circle arc.
    p, c, l = add_x_rectangle_loop(x, y, z, ly, lz)

    # Add hole loops to the list of loops defining the surface.
    ls = [l] + hole_loops

    # Add the surface.
    s = gmsh.model.geo.addPlaneSurface(ls)

    return p, c, l, s


def add_y_rectangle(x, y, z, lx, lz, hole_loops):

    # Add the corresponding circle arc.
    p, c, l = add_y_rectangle_loop(x, y, z, lx, lz)

    # Add hole loops to the list of loops defining the surface.
    ls = [l] + hole_loops

    # Add the surface.
    s = gmsh.model.geo.addPlaneSurface(ls)

    return p, c, l, s


def add_z_rectangle(x, y, z, lx, ly, hole_loops):

    # Add the corresponding circle arc.
    p, c, l = add_z_rectangle_loop(x, y, z, lx, ly)

    # Add hole loops to the list of loops defining the surface.
    ls = [l] + hole_loops

    # Add the surface.
    s = gmsh.model.geo.addPlaneSurface(ls)

    return p, c, l, s


def add_circle_arc(x, y, z, r):
    """Adds a circle arc to an open gmsh model.

    The circle arc is centered at [`x`, `y`, `z`], has radius `r` and is assumed
    to have a normal parallel to the z axis.
    """

    # First define the points required to draw the circle arc, i.e. the center
    # point (p1) and two pairs of diametrically opposed points (p2-p5).
    p1 = gmsh.model.geo.addPoint(x, y, z, 0)
    p2 = gmsh.model.geo.addPoint(x + r, y, z, 0)
    p3 = gmsh.model.geo.addPoint(x, y + r, z, 0)
    p4 = gmsh.model.geo.addPoint(x - r, y, z, 0)
    p5 = gmsh.model.geo.addPoint(x, y - r, z, 0)

    # Connect the multiple points into curves, e.g. c1 connects p2 to p3 with a
    # circle arc centered in p1.
    c1 = gmsh.model.geo.addCircleArc(p2, p1, p3)
    c2 = gmsh.model.geo.addCircleArc(p3, p1, p4)
    c3 = gmsh.model.geo.addCircleArc(p4, p1, p5)
    c4 = gmsh.model.geo.addCircleArc(p5, p1, p2)

    # Merge the multiple curves into a loop.
    l = gmsh.model.geo.addCurveLoop([c1, c2, c3, c4])

    return [p2, p3, p4, p5, p1], [c1, c2, c3, c4], l


def add_circle(x, y, z, r, hole_loops):
    """Adds a circle to an open gmsh model.

    The circle is centered at [`x`, `y`, `z`], has radius `r` and is assumed to
    have a normal parallel to the z axis.

    A circle differs from a circle arc in that it has its interior filled with a
    (plane) surface. This surface can have holes represented by the loops in
    `hole_loops`.
    """

    # Add the corresponding circle arc.
    p, c, l = add_circle_arc(x, y, z, r)

    # Add hole loops to the list of loops defining the surface.
    ls = [l] + hole_loops

    # Add the surface.
    s = gmsh.model.geo.addPlaneSurface(ls)

    return p, c, l, s


def add_cylinder_walls(p_base, c_base, p_top, c_top):
    """Adds cylinder walls (i.e. the curved surface) to an open gmsh model.

    The cylinder walls are built by merging the points and curves defining the
    bottom and top bases of the cylinder.

    More specifically, they are built in four steps, each focused on a curve
    defining a quarter of the base circle arcs.

    In each step, a line is created to connect a point in the bottom base with
    the corresponding point in the top base. Then, the sub-arc of the top base
    starting at that point is connected with the line. Another line is then
    added to connect the end point of the sub-arc in the top base to the
    corresponding point in the bottom base. Finally, the missing sub-arc in the
    bottom base is added (reversed relative to its definition, to ensure that
    the final set of curves defines a closed loop).
    """

    num_curves = len(c_base)

    # Loop through the steps, which number matches the number of curves in the
    # bases of the cylinder.
    for idx in range(num_curves):

        # Define the start and end indices.
        idx_start = idx
        idx_end = (idx + 1) % num_curves

        # Get the list of curves defining a quarter of the cylinder curved
        # surface:
        # 1. a line connecting the bottom and top bases;
        # 2. a sub-arc in the top base;
        # 3. a line connecting the top and bottom bases;
        # 4. a sub-arc in the bottom base (reversed relative to its definition,
        #    to ensure that the final set of curves defines a closed loop)-
        curves = [
            gmsh.model.geo.addLine(p_base[idx_start], p_top[idx_start]),
            c_top[idx],
            gmsh.model.geo.addLine(p_top[idx_end], p_base[idx_end]),
            -c_base[idx],
        ]

        loop = gmsh.model.geo.addCurveLoop(curves)
        gmsh.model.geo.addSurfaceFilling([loop])


def add_box(x, y, z, lx, ly, lz):
    gmsh.model.occ.addBox(
        x,
        y,
        z,
        lx,
        ly,
        lz,
    )


def add_cylinder(x, y, z, radius, height):
    gmsh.model.occ.addCylinder(
        x,
        y,
        z,
        0,
        0,
        height,
        radius,
    )


def convert_msh_to_obj_file(obj_file_path: str,
                            msh_file_path: str = TMP_MSH_FILE_PATH,
                            remove_msh_file: bool = True) -> None:
    """Converts msh to obj files using meshio."""

    # Read mesh using meshio.
    mesh = meshio.read(msh_file_path, file_format="gmsh")

    # Filter cell blocks to keep only those representing surfaces.
    cells = [block for block in mesh.cells if block.type == "triangle"]

    # Write mesh to obj file.
    meshio.write_points_cells(obj_file_path, mesh.points, cells)

    if remove_msh_file:
        os.remove(msh_file_path)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class InstanceGroup(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "machine_type",
            "image_name",
            "spot",
            "name",
            "disk_size_gb",
            "num_instances",
        }

        class properties:
            name = schemas.StrSchema
            machine_type = schemas.StrSchema
            image_name = schemas.StrSchema
            spot = schemas.BoolSchema
            disk_size_gb = schemas.IntSchema
            num_instances = schemas.IntSchema
            zone = schemas.StrSchema

            class resource_pool_id(
                    schemas.UUIDBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'uuid'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'resource_pool_id':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "name": name,
                "machine_type": machine_type,
                "image_name": image_name,
                "spot": spot,
                "disk_size_gb": disk_size_gb,
                "num_instances": num_instances,
                "zone": zone,
                "resource_pool_id": resource_pool_id,
            }

    machine_type: MetaOapg.properties.machine_type
    image_name: MetaOapg.properties.image_name
    spot: MetaOapg.properties.spot
    name: MetaOapg.properties.name
    disk_size_gb: MetaOapg.properties.disk_size_gb
    num_instances: MetaOapg.properties.num_instances

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["image_name"]
    ) -> MetaOapg.properties.image_name:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["disk_size_gb"]
    ) -> MetaOapg.properties.disk_size_gb:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_instances"]
    ) -> MetaOapg.properties.num_instances:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["zone"]
    ) -> MetaOapg.properties.zone:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["resource_pool_id"]
    ) -> MetaOapg.properties.resource_pool_id:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "machine_type",
        "image_name",
        "spot",
        "disk_size_gb",
        "num_instances",
        "zone",
        "resource_pool_id",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["image_name"]
    ) -> MetaOapg.properties.image_name:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["disk_size_gb"]
    ) -> MetaOapg.properties.disk_size_gb:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_instances"]
    ) -> MetaOapg.properties.num_instances:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["zone"]
    ) -> typing.Union[MetaOapg.properties.zone, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["resource_pool_id"]
    ) -> typing.Union[MetaOapg.properties.resource_pool_id, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "machine_type",
        "image_name",
        "spot",
        "disk_size_gb",
        "num_instances",
        "zone",
        "resource_pool_id",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        machine_type: typing.Union[
            MetaOapg.properties.machine_type,
            str,
        ],
        image_name: typing.Union[
            MetaOapg.properties.image_name,
            str,
        ],
        spot: typing.Union[
            MetaOapg.properties.spot,
            bool,
        ],
        name: typing.Union[
            MetaOapg.properties.name,
            str,
        ],
        disk_size_gb: typing.Union[
            MetaOapg.properties.disk_size_gb,
            decimal.Decimal,
            int,
        ],
        num_instances: typing.Union[
            MetaOapg.properties.num_instances,
            decimal.Decimal,
            int,
        ],
        zone: typing.Union[MetaOapg.properties.zone, str,
                           schemas.Unset] = schemas.unset,
        resource_pool_id: typing.Union[MetaOapg.properties.resource_pool_id,
                                       dict, frozendict.frozendict, str, date,
                                       datetime, uuid.UUID, int, float,
                                       decimal.Decimal, bool, None, list, tuple,
                                       bytes, io.FileIO, io.BufferedReader,
                                       schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'InstanceGroup':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            image_name=image_name,
            spot=spot,
            name=name,
            disk_size_gb=disk_size_gb,
            num_instances=num_instances,
            zone=zone,
            resource_pool_id=resource_pool_id,
            _configuration=_configuration,
            **kwargs,
        )



# pylint: disable=missing-docstring
from .misc import split_camel_case


def test_split_camel_case():
    assert split_camel_case("DamBreak") == ["Dam", "Break"]
    assert split_camel_case("CamelCamelCase") == ["Camel", "Camel", "Case"]
    assert split_camel_case("Camel2Camel2Case") == ["Camel2", "Camel2", "Case"]
    assert split_camel_case("getHTTPResponseCode") == [
        "get", "HTTP", "Response", "Code"
    ]
    assert split_camel_case("get2HTTPResponseCode") == [
        "get2", "HTTP", "Response", "Code"
    ]
    assert split_camel_case("HTTPResponseCode") == ["HTTP", "Response", "Code"]
    assert split_camel_case("HTTPResponseCodeXYZ") == [
        "HTTP", "Response", "Code", "XYZ"
    ]
    assert split_camel_case("") == []



"""Functions that use the API to compute eigenvalues.

These functions compute the eigenvalues and eigenvectors of
sparse metrices with the corresponding solvers.
"""

from typing import Tuple

import numpy as np
import scipy
from inductiva.api import invoke_api_from_fn_ptr
# pylint: disable=unused-argument


def eigs(matrix: scipy.sparse,
         num_eigenpairs: int) -> Tuple[np.ndarray, np.ndarray]:
    return invoke_api_from_fn_ptr(locals(), eigs)


# pylint: enable=unused-argument



import typing_extensions

from inductiva.client.paths import PathValues
from inductiva.client.apis.paths.tasks_submit import TasksSubmit
from inductiva.client.apis.paths.tasks_task_id_input import TasksTaskIdInput
from inductiva.client.apis.paths.tasks_task_id import TasksTaskId
from inductiva.client.apis.paths.tasks import Tasks
from inductiva.client.apis.paths.tasks_task_id_status import TasksTaskIdStatus
from inductiva.client.apis.paths.tasks_task_id_output import TasksTaskIdOutput
from inductiva.client.apis.paths.tasks_task_id_kill import TasksTaskIdKill
from inductiva.client.apis.paths.admin_users import AdminUsers
from inductiva.client.apis.paths.admin_users_username import AdminUsersUsername
from inductiva.client.apis.paths.admin_users_username_tasks import AdminUsersUsernameTasks
from inductiva.client.apis.paths.admin_tasks import AdminTasks
from inductiva.client.apis.paths.executers_register import ExecutersRegister
from inductiva.client.apis.paths.executers_pools import ExecutersPools
from inductiva.client.apis.paths.gcp_instances import GcpInstances
from inductiva.client.apis.paths.gcp_instances_group import GcpInstancesGroup
from inductiva.client.apis.paths.gcp_instances_price import GcpInstancesPrice

PathToApi = typing_extensions.TypedDict(
    'PathToApi', {
        PathValues.TASKS_SUBMIT: TasksSubmit,
        PathValues.TASKS_TASK_ID_INPUT: TasksTaskIdInput,
        PathValues.TASKS_TASK_ID: TasksTaskId,
        PathValues.TASKS: Tasks,
        PathValues.TASKS_TASK_ID_STATUS: TasksTaskIdStatus,
        PathValues.TASKS_TASK_ID_OUTPUT: TasksTaskIdOutput,
        PathValues.TASKS_TASK_ID_KILL: TasksTaskIdKill,
        PathValues.ADMIN_USERS: AdminUsers,
        PathValues.ADMIN_USERS_USERNAME: AdminUsersUsername,
        PathValues.ADMIN_USERS_USERNAME_TASKS: AdminUsersUsernameTasks,
        PathValues.ADMIN_TASKS: AdminTasks,
        PathValues.EXECUTERS_REGISTER: ExecutersRegister,
        PathValues.EXECUTERS_POOLS: ExecutersPools,
        PathValues.GCP_INSTANCES: GcpInstances,
        PathValues.GCP_INSTANCES_GROUP: GcpInstancesGroup,
        PathValues.GCP_INSTANCES_PRICE: GcpInstancesPrice,
    })

path_to_api = PathToApi({
    PathValues.TASKS_SUBMIT: TasksSubmit,
    PathValues.TASKS_TASK_ID_INPUT: TasksTaskIdInput,
    PathValues.TASKS_TASK_ID: TasksTaskId,
    PathValues.TASKS: Tasks,
    PathValues.TASKS_TASK_ID_STATUS: TasksTaskIdStatus,
    PathValues.TASKS_TASK_ID_OUTPUT: TasksTaskIdOutput,
    PathValues.TASKS_TASK_ID_KILL: TasksTaskIdKill,
    PathValues.ADMIN_USERS: AdminUsers,
    PathValues.ADMIN_USERS_USERNAME: AdminUsersUsername,
    PathValues.ADMIN_USERS_USERNAME_TASKS: AdminUsersUsernameTasks,
    PathValues.ADMIN_TASKS: AdminTasks,
    PathValues.EXECUTERS_REGISTER: ExecutersRegister,
    PathValues.EXECUTERS_POOLS: ExecutersPools,
    PathValues.GCP_INSTANCES: GcpInstances,
    PathValues.GCP_INSTANCES_GROUP: GcpInstancesGroup,
    PathValues.GCP_INSTANCES_PRICE: GcpInstancesPrice,
})



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskStatusCode(schemas.EnumBase, schemas.StrSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Possible task status codes.
    """

    class MetaOapg:
        enum_value_to_name = {
            "pending-input": "PENDINGINPUT",
            "submitted": "SUBMITTED",
            "started": "STARTED",
            "success": "SUCCESS",
            "failed": "FAILED",
            "pending-kill": "PENDINGKILL",
            "killed": "KILLED",
            "spot-instance-preempted": "SPOTINSTANCEPREEMPTED",
            "executer-terminated": "EXECUTERTERMINATED",
            "executer-failed": "EXECUTERFAILED",
        }

    @schemas.classproperty
    def PENDINGINPUT(cls):
        return cls("pending-input")

    @schemas.classproperty
    def SUBMITTED(cls):
        return cls("submitted")

    @schemas.classproperty
    def STARTED(cls):
        return cls("started")

    @schemas.classproperty
    def SUCCESS(cls):
        return cls("success")

    @schemas.classproperty
    def FAILED(cls):
        return cls("failed")

    @schemas.classproperty
    def PENDINGKILL(cls):
        return cls("pending-kill")

    @schemas.classproperty
    def KILLED(cls):
        return cls("killed")

    @schemas.classproperty
    def SPOTINSTANCEPREEMPTED(cls):
        return cls("spot-instance-preempted")

    @schemas.classproperty
    def EXECUTERTERMINATED(cls):
        return cls("executer-terminated")

    @schemas.classproperty
    def EXECUTERFAILED(cls):
        return cls("executer-failed")



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.apis.path_to_api import path_to_api

import enum


class PathValues(str, enum.Enum):
    TASKS_SUBMIT = "/tasks/submit"
    TASKS_TASK_ID_INPUT = "/tasks/{task_id}/input"
    TASKS_TASK_ID = "/tasks/{task_id}"
    TASKS = "/tasks"
    TASKS_TASK_ID_STATUS = "/tasks/{task_id}/status"
    TASKS_TASK_ID_OUTPUT = "/tasks/{task_id}/output"
    TASKS_TASK_ID_KILL = "/tasks/{task_id}/kill"
    ADMIN_USERS = "/admin/users"
    ADMIN_USERS_USERNAME = "/admin/users/{username}"
    ADMIN_USERS_USERNAME_TASKS = "/admin/users/{username}/tasks"
    ADMIN_TASKS = "/admin/tasks"
    EXECUTERS_REGISTER = "/executers/register"
    EXECUTERS_POOLS = "/executers/pools"
    GCP_INSTANCES = "/gcp_instances"
    GCP_INSTANCES_GROUP = "/gcp_instances/group"
    GCP_INSTANCES_PRICE = "/gcp_instances/price"
